#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from base64 import b64decode
import os
import sys

try:
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
    from cryptography.hazmat.primitives.serialization import load_pem_private_key

    HAS_CRYPTOGRAPHY = True
except ImportError:
    HAS_CRYPTOGRAPHY = False

try:
    import boto3

    HAS_BOTO3 = True
except ImportError:
    HAS_BOTO3 = False

from molecule import logger
from molecule.api import Driver

from molecule import util

LOG = logger.get_logger(__name__)


class EC2(Driver):
    """
    The class responsible for managing `EC2`_ instances.  `EC2`_
    is ``not`` the default driver used in Molecule.

    Molecule leverages Ansible's `ec2_module`_, by mapping variables from
    ``molecule.yml`` into ``create.yml`` and ``destroy.yml``.

    .. _`ec2_module`: https://docs.ansible.com/ansible/latest/ec2_module.html

    .. code-block:: yaml

        driver:
          name: ec2
        platforms:
          - name: instance

    Some configuration examples:

    .. code-block:: yaml

        driver:
          name: ec2
        platforms:
          - name: instance
            image: ami-0311dc90a352b25f4
            instance_type: t2.micro
            vpc_subnet_id: subnet-1cb17175

    If you don't know the AMI code or want to avoid hardcoding it:

    .. code-block:: yaml

        driver:
          name: ec2
        platforms:
          - name: instance
            image_owner: 099720109477
            image_name: ubuntu/images/hvm-ssd/ubuntu-bionic-18.04-amd64-server-20190320
            instance_type: t2.micro
            vpc_subnet_id: subnet-1cb17175

    Use wildcards for getting the latest image. For example, the latest Ubuntu bionic image:

    .. code-block:: yaml

        driver:
          name: ec2
        platforms:
          - name: instance
            image_owner: 099720109477
            image_name: ubuntu/images/hvm-ssd/ubuntu-bionic-18.04-amd64-server-*
            instance_type: t2.micro
            vpc_subnet_id: subnet-1cb17175

    Windows EC2 instances can be used as well:

    .. code-block:: yaml

        driver:
          name: ec2
        platforms:
          - name: instance
            image: ami-06a0d33fc8d328de0
            instance_type: t3a.medium
            vpc_subnet_id: subnet-1cb17175
            connection_options:
              sudo: False
              ansible_user: Administrator
              # Specify a password to override the automatic lookup
              # or omit to retrieve automatically
              # (requires boto3 & cryptography packages)
              # ansible_password: hunter2
              ansible_port: 5986
              ansible_connection: winrm
              ansible_winrm_scheme: https
              ansible_winrm_server_cert_validation: ignore

    .. code-block:: bash

        $ pip install 'molecule[ec2]'

    Change the options passed to the ssh client.

    .. code-block:: yaml

        driver:
          name: ec2
          ssh_connection_options:
            - '-o ControlPath=~/.ansible/cp/%r@%h-%p'

    .. important::

        Molecule does not merge lists, when overriding the developer must
        provide all options.

    Provide a list of files Molecule will preserve, relative to the scenario
    ephemeral directory, after any ``destroy`` subcommand execution.

    .. code-block:: yaml

        driver:
          name: ec2
          safe_files:
            - foo

    .. _`EC2`: https://aws.amazon.com/ec2/
    """  # noqa

    def __init__(self, config=None):
        super(EC2, self).__init__(config)
        self._name = "ec2"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def login_cmd_template(self):
        if self._config.command_args.get("host"):
            hostname = self._config.command_args["host"]
        elif len(self._config.platforms.instances) == 1:
            hostname = self._config.platforms.instances[0]["name"]
        else:
            LOG.error("Please specify instance via '--host'")
            sys.exit(1)

        ansible_connection_options = self.ansible_connection_options(hostname)
        if ansible_connection_options.get("ansible_connection") == "winrm":
            return (
                "xfreerdp "
                '"/u:%s" '
                '"/p:%s" '
                "/v:%s "
                "/cert-tofu "
                "+clipboard "
                "/grab-keyboard"
                % (
                    ansible_connection_options["ansible_user"],
                    ansible_connection_options["ansible_password"],
                    ansible_connection_options["ansible_host"],
                )
            )

        else:  # normal ssh connection
            connection_options = " ".join(self.ssh_connection_options)

            return (
                "ssh {{address}} "
                "-l {{user}} "
                "-p {{port}} "
                "-i {{identity_file}} "
                "{}"
            ).format(connection_options)

    @property
    def default_safe_files(self):
        return [self.instance_config]

    @property
    def default_ssh_connection_options(self):
        return self._get_ssh_connection_options()

    def login_options(self, instance_name):
        d = {"instance": instance_name}

        return util.merge_dicts(d, self._get_instance_config(instance_name))

    def ansible_connection_options(self, instance_name):
        try:
            d = self._get_instance_config(instance_name)
            plat_conn_opts = next(
                (
                    item
                    for item in self._config.config.get("platforms", [])
                    if item["name"] == instance_name
                ),
                {},
            ).get("connection_options", {})
            conn_opts = util.merge_dicts(
                {
                    "ansible_user": d["user"],
                    "ansible_host": d["address"],
                    "ansible_port": d["port"],
                    "ansible_private_key_file": d["identity_file"],
                    "connection": "ssh",
                    "ansible_ssh_common_args": " ".join(self.ssh_connection_options),
                },
                plat_conn_opts,
            )
            if conn_opts.get("ansible_connection") == "winrm" and (
                not conn_opts.get("ansible_password")
            ):
                conn_opts["ansible_password"] = self._get_windows_instance_pass(
                    d["instance_ids"][0], d["identity_file"]
                )
            return conn_opts
        except StopIteration:
            return {}
        except IOError:
            # Instance has yet to be provisioned , therefore the
            # instance_config is not on disk.
            return {}

    def _get_instance_config(self, instance_name):
        instance_config_dict = util.safe_load_file(self._config.driver.instance_config)

        return next(
            item for item in instance_config_dict if item["instance"] == instance_name
        )

    def _get_windows_instance_pass(self, instance_id, key_file):
        if not HAS_BOTO3:
            LOG.error("boto3 required when using Windows instances")
            sys.exit(1)
        if not HAS_CRYPTOGRAPHY:
            LOG.error("cryptography package required when using Windows instances")
            sys.exit(1)
        ec2_client = boto3.client("ec2")
        data_response = ec2_client.get_password_data(InstanceId=instance_id)
        decoded = b64decode(data_response["PasswordData"])
        with open(key_file, "rb") as f:
            key = load_pem_private_key(f.read(), None, default_backend())
        return key.decrypt(decoded, PKCS1v15()).decode("utf-8")

    def sanity_checks(self):
        # FIXME(decentral1se): Implement sanity checks
        pass

    def template_dir(self):
        """Return path to its own cookiecutterm templates. It is used by init
        command in order to figure out where to load the templates from.
        """
        return os.path.join(os.path.dirname(__file__), "cookiecutter")



"""Functional tests."""
import os
import pathlib
import shutil
import subprocess
import pytest

from molecule import logger
from molecule.test.conftest import change_dir_to
from molecule.util import run_command

LOG = logger.get_logger(__name__)


def format_result(result: subprocess.CompletedProcess):
    """Return friendly representation of completed process run."""
    return (
        f"RC: {result.returncode}\n"
        + f"STDOUT: {result.stdout}\n"
        + f"STDERR: {result.stderr}"
    )


@pytest.mark.skip(reason="broken, fix welcomed")
def test_command_init_and_test_scenario(tmp_path: pathlib.Path, DRIVER: str) -> None:
    """Verify that init scenario works."""
    shutil.rmtree(tmp_path, ignore_errors=True)
    tmp_path.mkdir(exist_ok=True)

    scenario_name = "default"

    with change_dir_to(tmp_path):

        scenario_directory = tmp_path / "molecule" / scenario_name
        cmd = [
            "molecule",
            "init",
            "scenario",
            "--driver-name",
            DRIVER,
        ]
        result = run_command(cmd)
        assert result.returncode == 0

        assert scenario_directory.exists()

        # run molecule reset as this may clean some leftovers from other
        # test runs and also ensure that reset works.
        result = run_command(["molecule", "reset"])  # default scenario
        assert result.returncode == 0

        result = run_command(["molecule", "reset", "-s", scenario_name])
        assert result.returncode == 0

        cmd = ["molecule", "--debug", "test", "-s", scenario_name]
        result = run_command(cmd)
        assert result.returncode == 0


@pytest.mark.skip(reason="broken, fix welcomed")
def test_command_static_scenario() -> None:
    """Validate that the scenario we included with code still works."""
    cmd = ["molecule", "test"]

    result = run_command(cmd)
    assert result.returncode == 0


@pytest.mark.skip(reason="broken, fix welcomed")
def test_dockerfile_with_context() -> None:
    """Verify that Dockerfile.j2 with context works."""
    with change_dir_to("test/docker/scenarios/with-context"):
        cmd = ["molecule", "--debug", "test"]
        result = run_command(cmd)
        assert result.returncode == 0


@pytest.mark.skip(reason="broken, fix welcomed")
def test_env_substitution() -> None:
    """Verify that env variables in molecule.yml are replaced properly."""
    os.environ["MOLECULE_ROLE_IMAGE"] = "debian:bullseye"
    with change_dir_to("test/docker/scenarios/env-substitution"):
        cmd = ["molecule", "--debug", "test"]
        result = run_command(cmd)
        assert result.returncode == 0



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#  Copyright (c) 2018 Red Hat, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

import platform
import shutil
import pytest
import os
import vagrant

from molecule import util
from molecule import logger
from molecule.scenario import ephemeral_directory
from molecule.util import run_command
from molecule.test.conftest import change_dir_to

LOG = logger.get_logger(__name__)


def is_vagrant_supported() -> bool:
    """Return True if vagrant is installed and current platform is supported."""
    if not shutil.which("vagrant"):
        return False
    if platform.machine() == "arm64" and platform.system() == "Darwin":
        return False
    return True


@pytest.mark.skipif(
    not is_vagrant_supported(), reason="vagrant not supported on this machine"
)
def test_command_init_scenario(temp_dir):
    with change_dir_to(temp_dir):
        os.makedirs(os.path.join(temp_dir, "molecule", "default"))
        scenario_directory = os.path.join(temp_dir, "molecule", "test-scenario")
        cmd = [
            "molecule",
            "init",
            "scenario",
            "test-scenario",
            "--driver-name",
            "vagrant",
        ]
        result = run_command(cmd)
        assert result.returncode == 0

        assert os.path.isdir(scenario_directory)
        confpath = os.path.join(scenario_directory, "molecule.yml")
        conf = util.safe_load_file(confpath)
        env = os.environ
        if "TESTBOX" in env:
            conf["platforms"][0]["box"] = env["TESTBOX"]
        if "vagrant-libvirt" in list(
            map(lambda x: x.name, vagrant.Vagrant().plugin_list())
        ):
            conf["driver"]["provider"] = {"name": "libvirt"}
        util.write_file(confpath, util.safe_dump(conf))

        cmd = ["molecule", "--debug", "test", "-s", "test-scenario"]
        result = run_command(cmd)
        assert result.returncode == 0


@pytest.mark.skipif(
    not is_vagrant_supported(), reason="vagrant not supported on this machine"
)
def test_invalid_settings(temp_dir):

    scenario_directory = os.path.join(
        os.path.dirname(util.abs_path(__file__)), os.path.pardir, "scenarios"
    )

    with change_dir_to(scenario_directory):
        cmd = ["molecule", "create", "--scenario-name", "invalid"]
        result = run_command(cmd)
        assert result.returncode == 2

        assert "Failed to validate generated Vagrantfile" in result.stdout


@pytest.mark.skipif(
    not is_vagrant_supported(), reason="vagrant not supported on this machine"
)
@pytest.mark.parametrize(
    "scenario",
    [
        ("vagrant_root"),
        ("config_options"),
        ("provider_config_options"),
        ("default"),
        ("default-compat"),
        ("network"),
        ("hostname"),
    ],
)
def test_vagrant_root(temp_dir, scenario):

    scenario_directory = os.path.join(
        os.path.dirname(util.abs_path(__file__)), os.path.pardir, "scenarios"
    )

    with change_dir_to(scenario_directory):
        cmd = ["molecule", "test", "--scenario-name", scenario]
        result = run_command(cmd)
        assert result.returncode == 0


@pytest.mark.skipif(
    not is_vagrant_supported(), reason="vagrant not supported on this machine"
)
def test_multi_node(temp_dir):

    scenario_directory = os.path.join(
        os.path.dirname(util.abs_path(__file__)), os.path.pardir, "scenarios"
    )

    with change_dir_to(scenario_directory):
        cmd = ["molecule", "test", "--scenario-name", "multi-node"]
        result = run_command(cmd)
        assert result.returncode == 0

    molecule_eph_directory = ephemeral_directory()
    vagrantfile = os.path.join(
        molecule_eph_directory, "scenarios", "multi-node", "Vagrantfile"
    )
    with open(vagrantfile) as f:
        content = f.read()
        assert "instance-1" in content
        assert "instance-2" in content



#!/usr/bin/python
# -*- coding: utf-8 -*-

#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.


from __future__ import absolute_import, division, print_function

__metaclass__ = type

from ansible.module_utils.basic import AnsibleModule
import contextlib
import datetime
import io
import os
import subprocess
import sys

import molecule
import molecule.util

try:
    import vagrant
except ImportError:
    sys.exit("ERROR: Driver missing, install python-vagrant.")
ANSIBLE_METADATA = {
    "metadata_version": "0.1",
    "status": ["preview"],
    "supported_by": "community",
}

DOCUMENTATION = """
---
module: vagrant
short_description: Manage Vagrant instances
description:
  - Manage the life cycle of Vagrant instances.
version_added: 2.0
author:
  - Cisco Systems, Inc.
options:
  instances:
    description:
      - list of instances to handle
    required: False
    default: []
  instance_name:
    description:
      - Assign a name to a new instance or match an existing instance.
    required: False
    default: None
  instance_interfaces:
    description:
      - Assign interfaces to a new instance.
    required: False
    default: []
  instance_raw_config_args:
    description:
      - Additional Vagrant options not explicitly exposed by this module.
    required: False
    default: None
  config_options:
    description:
      - Additional config options not explicitly exposed by this module.
    required: False
    default: {}
  platform_box:
    description:
      - Name of Vagrant box.
    required: True
    default: None
  platform_box_version:
    description:
      - Explicit version of Vagrant box to use.
    required: False
    default: None
  platform_box_url:
    description:
      - The URL to a Vagrant box.
    required: False
    default: None
  platform_box_download_checksum:
    description:
      - The checksum of the box specified by platform_box_url.
    required: False
    default: None
  platform_box_download_checksum_type:
    description:
      - The type of checksum specified by platform_box_download_checksum (if any).
    required: False
    default: None
  provider_name:
    description:
      - Name of the Vagrant provider to use.
    required: False
    default: virtualbox
  provider_memory:
    description:
      - Amount of memory to allocate to the instance.
    required: False
    default: 512
  provider_cpus:
    description:
      - Number of CPUs to allocate to the instance.
    required: False
    default: 2
  provider_options:
    description:
      - Additional provider options not explicitly exposed by this module.
    required: False
    default: {}
  provider_override_args:
    description:
      - Additional override options not explicitly exposed by this module.
    required: False
    default: None
  provider_raw_config_args:
    description:
      - Additional Vagrant options not explicitly exposed by this module.
    required: False
    default: None
  force_stop:
    description:
      - Force halt the instance, then destroy the instance.
    required: False
    default: False
  state:
    description:
      - The desired state of the instance.
    required: True
    choices: ['up', 'halt', 'destroy']
    default: None
  workdir:
    description:
      - vagrant working directory
    required: False
    default: content of MOLECULE_EPHEMERAL_DIRECTORY environment variable
  default_box:
    description:
      - Default vagrant box to use when not specified in instance configuration
    required: False
    default: None
  cachier:
    description:
      - Cachier scope configuration. Can be "machine" or "box". Any other value
        will disable cachier.
    required: False
    default: "machine"
  parallel:
    description:
      - Enable or disable parallelism when bringing up the VMs by
        setting VAGRANT_NO_PARALLEL environment variable.
    required: False
    default: True

requirements:
    - python >= 2.6
    - python-vagrant
    - vagrant
"""

EXAMPLES = """
See doc/source/configuration.rst
"""

VAGRANTFILE_TEMPLATE = """
{%- macro ruby_format(value) -%}
  {%- if value is boolean -%}
    {{ value | string | lower }}
  {%- elif value is string -%}
    {# "Bug" compat. To be removed later #}
    {%- if value[0] == value[-1] and value.startswith(("'", '"')) -%}
    {{ value }}
    {%- else -%}
    "{{ value }}"
    {%- endif -%}
  {%- else -%}
    {{ value }}
  {%- endif -%}
{%- endmacro -%}

{%- macro dict2args(dictionary) -%}
  {% set sep = joiner(", ") %}
  {%- for key, value in dictionary.items() -%}
    {{ sep() }}{{ key }}: {{ ruby_format(value) }}
  {%- endfor -%}
{%- endmacro -%}

Vagrant.configure('2') do |config|
  if Vagrant.has_plugin?('vagrant-cachier')
    {% if cachier is not none and cachier in [ "machine", "box" ] %}
    config.cache.scope = '{{ cachier }}'
    {% else %}
    config.cache.disable!
    {% endif %}
  end

{% for instance in instances %}
  config.vm.define "{{ instance.name }}" do |c|
    ##
    # Box definition
    ##
    c.vm.box = "{{ instance.box }}"
    {{ 'c.vm.box_version = "{}"'.format(instance.box_version) if instance.box_version }}
    {{ 'c.vm.box_url = "{}"'.format(instance.box_url) if instance.box_url }}
    {{ 'c.vm.box_download_checksum = "{}"'.format(instance.box_download_checksum) if instance.box_download_checksum }}
    {{ 'c.vm.box_download_checksum_type = "{}"'.format(instance.box_download_checksum_type) if instance.box_download_checksum_type }}

    ##
    # Config options
    ##
    {% if instance.config_options['synced_folder'] is sameas false %}
    c.vm.synced_folder ".", "/vagrant", disabled: true
    {% endif %}

    {% for k,v in instance.config_options.items() %}
    {% if k not in ['synced_folder', 'cachier'] %}c.{{ k }} = {{ ruby_format(v) }}{% endif %}
    {% endfor %}

    {% if instance.hostname is boolean and instance.hostname is sameas false %}
    # c.vm.hostname not set
    {% elif instance.hostname is string %}
    c.vm.hostname = "{{ instance.hostname }}"
    {% else %}
    c.vm.hostname = "{{ instance.name }}"
    {% endif %}

    ##
    # Network
    ##
    {% for n in instance.networks %}
    c.vm.network "{{ n.name }}", {{ dict2args(n.options) }}
    {% endfor %}

    ##
    # instance_raw_config_args
    ##
    {% if instance.instance_raw_config_args is not none %}{% for arg in instance.instance_raw_config_args -%}
    c.{{ arg }}
    {% endfor %}{% endif %}

    ##
    # Provider
    ##
    c.vm.provider "{{ instance.provider }}" do |{{ instance.provider | lower }}, override|
      {% if instance.provider == "vsphere" %}
      {{ instance.provider | lower }}.memory_mb = {{ instance.memory }}
      {{ instance.provider | lower }}.cpu_count = {{ instance.cpus }}
      {% elif instance.provider.startswith('vmware_') %}
      {{ instance.provider | lower }}.vmx['memsize'] = {{ instance.memory }}
      {{ instance.provider | lower }}.vmx['numvcpus'] = {{ instance.cpus }}
      {% else %}
      {{ instance.provider | lower }}.memory = {{ instance.memory }}
      {{ instance.provider | lower }}.cpus = {{ instance.cpus }}
      {% endif %}

      {% for option, value in instance.provider_options.items() %}
      {{ instance.provider | lower }}.{{ option }} = {{ ruby_format(value) }}
      {% endfor %}

      {% if instance.provider_raw_config_args is not none %}
        {% for arg in instance.provider_raw_config_args %}
      {{ instance.provider | lower }}.{{ arg }}
        {% endfor %}
      {% endif %}

      {% if instance.provider_override_args is not none %}
        {% for arg in instance.provider_override_args -%}
      override.{{ arg }}
        {% endfor %}
      {% endif %}

      {% if instance.provider == 'virtualbox' %}
      {% if 'linked_clone' not in instance.provider_options %}
      virtualbox.linked_clone = true
      {% endif %}
      {% endif %}
      {% if instance.provider == 'libvirt' %}
        {% if no_kvm is sameas true and 'driver' not in instance.provider_options %}
      libvirt.driver='qemu'
        {% endif %}
        {% set libvirt_use_qemu = no_kvm %}
        {% if 'driver' in instance.provider_options and 'qemu' in instance.provider_options['driver'] %}
          {% set libvirt_use_qemu = true %}
        {% endif %}
        {% if libvirt_use_qemu is sameas true %}
          {% if 'cpu_mode' not in instance.provider_options %}
      # When using qemu instead of kvm, some libvirt systems
      # will use EPYC as vCPU model inside the new VM.
      # This will lead to process hang with ssh-keygen -A on alpine.
      # Not sure were the bug is (libvirt or qemu or alpine or all of them) but using QEMU64
      # cpu model will work around this. Hopefully, by checking 'cpu_mode' option, it will
      # allow people to override the model to use.
      libvirt.cpu_mode = 'custom'
      libvirt.cpu_model = 'qemu64'
          {% endif %}
        {% endif %}
      {% endif %}
    end
  end
{% endfor %}
end
""".strip()  # noqa

RETURN = r"""
rc:
    description: The command return code (0 means success)
    returned: always
    type: int
cmd:
    description: The command executed by the task
    returned: always
    type: str
stdout:
    description: The command standard output
    returned: changed
    type: str
stderr:
    description: Output on stderr
    returned: changed
    type: str
"""


class VagrantClient(object):
    def __init__(self, module):
        self._module = module
        self.provision = self._module.params["provision"]
        self.cachier = self._module.params["cachier"]

        # compat
        if self._module.params["instance_name"] is not None:
            self._module.warn(
                "Please convert your playbook to use the instances parameter. Compat layer will be removed later."
            )
            self.instances = [
                {
                    "name": self._module.params["instance_name"],
                    "interfaces": self._module.params["instance_interfaces"],
                    "instance_raw_config_args": self._module.params[
                        "instance_raw_config_args"
                    ],
                    "config_options": self._module.params["config_options"],
                    "box": self._module.params["platform_box"],
                    "box_version": self._module.params["platform_box_version"],
                    "box_url": self._module.params["platform_box_url"],
                    "box_download_checksum": self._module.params[
                        "platform_box_download_checksum"
                    ],
                    "box_download_checksum_type": self._module.params[
                        "platform_box_download_checksum_type"
                    ],
                    "memory": self._module.params["provider_memory"],
                    "cpus": self._module.params["provider_cpus"],
                    "provider_options": self._module.params["provider_options"],
                    "provider_override_args": self._module.params[
                        "provider_override_args"
                    ],
                    "provider_raw_config_args": self._module.params[
                        "provider_raw_config_args"
                    ],
                }
            ]
        else:
            self.instances = self._module.params["instances"]

        self._config = self._get_config()
        self._vagrantfile = self._config["vagrantfile"]
        self._vagrant = self._get_vagrant()
        self._write_configs()
        self._has_error = None
        self._datetime = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.result = {}

    @contextlib.contextmanager
    def stdout_cm(self):
        """Redirect the stdout to a log file."""
        with open(self._get_stdout_log(), "a+") as fh:
            msg = "### {} ###\n".format(self._datetime)
            fh.write(msg)
            fh.flush()

            yield fh

    @contextlib.contextmanager
    def stderr_cm(self):
        """Redirect the stderr to a log file."""
        with open(self._get_stderr_log(), "a+") as fh:
            msg = "### {} ###\n".format(self._datetime)
            fh.write(msg)
            fh.flush()

            try:
                yield fh
            except subprocess.CalledProcessError as e:
                self._has_error = True
                # msg = "CMD: {} returned {}\n{}".format(
                #     e.cmd, e.returncode, e.output or ""
                # )
                self.result["cmd"] = e.cmd
                self.result["rc"] = e.returncode
                self.result["stderr"] = e.output or ""

                # fh.write(msg)
                # raise
            except Exception as e:
                self._has_error = True
                if hasattr(e, "message"):
                    fh.write(e.message)
                else:
                    fh.write(e)
                fh.flush()
                raise

    def up(self):
        changed = False
        if self._running() != len(self.instances):
            changed = True
            provision = self.provision
            try:
                self._vagrant.up(provision=provision)
            except Exception:
                # NOTE(retr0h): Ignore the exception since python-vagrant
                # passes the actual error as a no-argument ContextManager.
                pass

        # NOTE(retr0h): Ansible wants only one module return `fail_json`
        # or `exit_json`.
        if not self._has_error:
            # compat
            if self._module.params["instance_name"] is not None:
                self._module.exit_json(
                    changed=changed, log=self._get_stdout_log(), **self._conf()[0]
                )
            self._module.exit_json(
                changed=changed, log=self._get_stdout_log(), results=self._conf()
            )

        msg = "Failed to start the VM(s): See log file '{}'".format(
            self._get_stderr_log()
        )
        with io.open(self._get_stderr_log(), "r", encoding="utf-8") as f:
            self.result["stderr"] = f.read()
        self._module.fail_json(msg=msg, **self.result)

    def destroy(self):
        changed = False
        if self._created() > 0:
            changed = True
            if self._module.params["force_stop"]:
                self._vagrant.halt(force=True)
            self._vagrant.destroy()

        self._module.exit_json(changed=changed)

    def halt(self):
        changed = False
        if self._running() > 0:
            changed = True
            self._vagrant.halt(force=self._module.params["force_stop"])

        self._module.exit_json(changed=changed)

    def _conf_instance(self, instance_name):
        try:
            return self._vagrant.conf(vm_name=instance_name)
        except Exception:
            msg = "Failed to get vagrant config for {}: See log file '{}'".format(
                instance_name, self._get_stderr_log()
            )
            with io.open(self._get_stderr_log(), "r", encoding="utf-8") as f:
                self.result["stderr"] = f.read()
                self._module.fail_json(msg=msg, **self.result)

    def _status_instance(self, instance_name):
        try:
            s = self._vagrant.status(vm_name=instance_name)[0]

            return {"name": s.name, "state": s.state, "provider": s.provider}
        except Exception:
            msg = "Failed to get status for {}: See log file '{}'".format(
                instance_name, self._get_stderr_log()
            )
            with io.open(self._get_stderr_log(), "r", encoding="utf-8") as f:
                self.result["stderr"] = f.read()
                self._module.fail_json(msg=msg, **self.result)

    def _conf(self):
        conf = []

        for i in self.instances:
            instance_name = i["name"]
            c = self._conf_instance(instance_name)
            if c:
                conf.append(c)

        return conf

    def _status(self):
        vms_status = []

        for i in self.instances:
            instance_name = i["name"]
            s = self._status_instance(instance_name)
            if s:
                vms_status.append(s)

        return vms_status

    def _created(self):
        status = self._status()
        if len(status) == 0:
            return 0

        count = sum(map(lambda s: s["state"] == "not_created", status))
        return len(status) - count

    def _running(self):
        status = self._status()
        if len(status) == 0:
            return 0

        count = sum(map(lambda s: s["state"] == "running", status))
        return count

    def _get_config(self):
        conf = dict()
        conf["workdir"] = os.getenv("MOLECULE_EPHEMERAL_DIRECTORY")
        if self._module.params["workdir"] is not None:
            conf["workdir"] = self._module.params["workdir"]
        if conf["workdir"] is None:
            self._module.fail_json(
                msg="Either workdir parameter or MOLECULE_EPHEMERAL_DIRECTORY env variable has to be set"
            )
        conf["vagrantfile"] = os.path.join(conf["workdir"], "Vagrantfile")
        return conf

    def _write_vagrantfile(self):
        instances = self._get_vagrant_config_dict()
        template = molecule.util.render_template(
            VAGRANTFILE_TEMPLATE,
            instances=instances,
            cachier=self.cachier,
            no_kvm=not os.path.exists("/dev/kvm"),
        )
        molecule.util.write_file(self._vagrantfile, template)

    def _write_configs(self):
        self._write_vagrantfile()
        try:
            self._vagrant.validate(self._config["workdir"])
        except subprocess.CalledProcessError as e:
            self._module.fail_json(
                msg=f"Failed to validate generated Vagrantfile: {e.stderr}"
            )

    def _get_vagrant(self):
        vagrant_env = os.environ.copy()
        if self._module.params["parallel"] is False:
            vagrant_env["VAGRANT_NO_PARALLEL"] = "1"
        v = vagrant.Vagrant(
            out_cm=self.stdout_cm,
            err_cm=self.stderr_cm,
            root=self._config["workdir"],
            env=vagrant_env,
        )

        return v

    def _get_instance_vagrant_config_dict(self, instance):

        checksum = instance.get("box_download_checksum")
        checksum_type = instance.get("box_download_checksum_type")
        if bool(checksum) ^ bool(checksum_type):
            self._module.fail_json(
                msg="box_download_checksum and box_download_checksum_type must be used together"
            )

        networks = []
        if "interfaces" in instance:
            for iface in instance["interfaces"]:
                net = dict()
                net["name"] = iface["network_name"]
                iface.pop("network_name")
                net["options"] = iface
                networks.append(net)

        # compat
        provision = instance.get("provision")
        if provision is not None:
            self.provision = self.provision or provision
            self._module.warn(
                "Please convert your molecule.yml to move provision parameter to driver:. Compat layer will be removed later."
            )
        d = {
            "name": instance.get("name"),
            "hostname": instance.get("hostname", instance.get("name")),
            "memory": instance.get("memory", 512),
            "cpus": instance.get("cpus", 2),
            "networks": networks,
            "instance_raw_config_args": instance.get("instance_raw_config_args", None),
            "config_options": {
                # NOTE(retr0h): `synced_folder` does not represent the
                # actual key used by Vagrant.  Is used as a flag to
                # simply enable/disable shared folder.
                "synced_folder": False,
                "ssh.insert_key": True,
            },
            "box": instance.get("box", self._module.params["default_box"]),
            "box_version": instance.get("box_version"),
            "box_url": instance.get("box_url"),
            "box_download_checksum": checksum,
            "box_download_checksum_type": checksum_type,
            "provider": self._module.params["provider_name"],
            "provider_options": {},
            "provider_raw_config_args": instance.get("provider_raw_config_args", None),
            "provider_override_args": instance.get("provider_override_args", None),
        }

        d["config_options"].update(
            molecule.util.merge_dicts(
                d["config_options"], instance.get("config_options", {})
            )
        )
        if "cachier" in d["config_options"]:
            self.cachier = d["config_options"]["cachier"]
            self._module.warn(
                "Please convert your molecule.yml to move cachier parameter to driver:. Compat layer will be removed later."
            )

        d["provider_options"].update(
            molecule.util.merge_dicts(
                d["provider_options"], instance.get("provider_options", {})
            )
        )

        return d

    def _get_vagrant_config_dict(self):
        config_list = []
        for instance in self.instances:
            config_list.append(self._get_instance_vagrant_config_dict(instance))
        return config_list

    def _get_stdout_log(self):
        return self._get_vagrant_log("out")

    def _get_stderr_log(self):
        return self._get_vagrant_log("err")

    def _get_vagrant_log(self, __type):
        return os.path.join(self._config["workdir"], "vagrant.{}".format(__type))


def main():
    module = AnsibleModule(
        argument_spec=dict(
            instances=dict(type="list", required=False),
            instance_name=dict(type="str", required=False, default=None),
            instance_interfaces=dict(type="list", default=[]),
            instance_raw_config_args=dict(type="list", default=None),
            config_options=dict(type="dict", default={}),
            platform_box=dict(type="str", required=False),
            platform_box_version=dict(type="str"),
            platform_box_url=dict(type="str"),
            platform_box_download_checksum=dict(type="str"),
            platform_box_download_checksum_type=dict(type="str"),
            provider_memory=dict(type="int", default=512),
            provider_cpus=dict(type="int", default=2),
            provider_options=dict(type="dict", default={}),
            provider_override_args=dict(type="list", default=None),
            provider_raw_config_args=dict(type="list", default=None),
            provider_name=dict(type="str", default="virtualbox"),
            default_box=dict(type="str", default=None),
            provision=dict(type="bool", default=False),
            force_stop=dict(type="bool", default=False),
            cachier=dict(type="str", default="machine"),
            state=dict(type="str", default="up", choices=["up", "destroy", "halt"]),
            workdir=dict(type="str"),
            parallel=dict(type="bool", default=True),
        ),
        required_together=[
            ("platform_box_download_checksum", "platform_box_download_checksum_type"),
            ("instances", "default_box"),
        ],
        supports_check_mode=False,
    )

    if not (bool(module.params["instances"]) ^ bool(module.params["instance_name"])):
        module.fail_json(
            msg="Either instances or instance_name parameters should be used and not at the same time"
        )

    v = VagrantClient(module)

    if module.params["state"] == "up":
        v.up()

    if module.params["state"] == "destroy":
        v.destroy()

    if module.params["state"] == "halt":
        v.halt()

    module.fail_json(msg="Unknown error", **v.result)


if __name__ == "__main__":
    main()



import contextlib
import os
import random
import string

import pytest
from molecule import config, logger, util
from molecule.scenario import ephemeral_directory

LOG = logger.get_logger(__name__)


@pytest.helpers.register
def run_command(cmd, env=os.environ, log=True):
    if cmd.__class__.__name__ == "Command":
        if log:
            cmd = _rebake_command(cmd, env)
        cmd = cmd.bake(_truncate_exc=False)
    return util.run_command(cmd, env=env)


def _rebake_command(cmd, env, out=LOG.info, err=LOG.error):
    return cmd.bake(_env=env, _out=out, _err=err)


@pytest.fixture
def random_string(length=5):
    return "".join((random.choice(string.ascii_uppercase) for _ in range(length)))


@contextlib.contextmanager
def change_dir_to(dir_name):
    cwd = os.getcwd()
    os.chdir(dir_name)
    yield
    os.chdir(cwd)


@pytest.fixture
def temp_dir(tmpdir, random_string, request):
    directory = tmpdir.mkdir(random_string)

    with change_dir_to(directory.strpath):
        yield directory


@pytest.fixture
def resources_folder_path():
    resources_folder_path = os.path.join(os.path.dirname(__file__), "resources")
    return resources_folder_path


@pytest.helpers.register
def molecule_project_directory():
    return os.getcwd()


@pytest.helpers.register
def molecule_directory():
    return config.molecule_directory(molecule_project_directory())


@pytest.helpers.register
def molecule_scenario_directory():
    return os.path.join(molecule_directory(), "default")


@pytest.helpers.register
def molecule_file():
    return get_molecule_file(molecule_scenario_directory())


@pytest.helpers.register
def get_molecule_file(path):
    return config.molecule_file(path)


@pytest.helpers.register
def molecule_ephemeral_directory(_fixture_uuid):
    project_directory = "test-project-{}".format(_fixture_uuid)
    scenario_name = "test-instance"

    return ephemeral_directory(
        os.path.join("molecule_test", project_directory, scenario_name)
    )



"""Unit tests."""
from molecule import api

from molecule_plugins.podman.driver import Podman


def test_driver_is_detected():
    """Asserts that molecule recognizes the driver."""
    assert any(str(d) == "podman" for d in api.drivers())


def test_driver_initializes_without_podman_executable(monkeypatch):
    """Make sure we can initiaize driver without having an executable present."""
    monkeypatch.setenv("MOLECULE_PODMAN_EXECUTABLE", "bad-executable")
    Podman()



#!/usr/bin/env python

# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import copy
import datetime
import json
import time
import argparse

# PyCrypto library: https://pypi.python.org/pypi/pycrypto
from Crypto.Cipher import PKCS1_OAEP
from Crypto.PublicKey import RSA
from Crypto.Util.number import long_to_bytes

# Google API Client Library for Python:
# https://developers.google.com/api-client-library/python/start/get_started
import google.auth
from googleapiclient.discovery import build


def GetCompute():
    """Get a compute object for communicating with the Compute Engine API."""
    credentials, project = google.auth.default()
    compute = build("compute", "v1", credentials=credentials)
    return compute


def GetInstance(compute, instance, zone, project):
    """Get the data for a Google Compute Engine instance."""
    cmd = compute.instances().get(instance=instance, project=project, zone=zone)
    return cmd.execute()


def GetKey():
    """Get an RSA key for encryption."""
    # This uses the PyCrypto library
    key = RSA.generate(2048)
    return key


def GetModulusExponentInBase64(key):
    """Return the public modulus and exponent for the key in bas64 encoding."""
    mod = long_to_bytes(key.n)
    exp = long_to_bytes(key.e)

    modulus = base64.b64encode(mod)
    exponent = base64.b64encode(exp)

    return modulus, exponent


def GetExpirationTimeString():
    """Return an RFC3339 UTC timestamp for 5 minutes from now."""
    utc_now = datetime.datetime.utcnow()
    # These metadata entries are one-time-use, so the expiration time does
    # not need to be very far in the future. In fact, one minute would
    # generally be sufficient. Five minutes allows for minor variations
    # between the time on the client and the time on the server.
    expire_time = utc_now + datetime.timedelta(minutes=5)
    return expire_time.strftime("%Y-%m-%dT%H:%M:%SZ")


def GetJsonString(user, modulus, exponent, email):
    """Return the JSON string object that represents the windows-keys entry."""

    converted_modulus = modulus.decode("utf-8")
    converted_exponent = exponent.decode("utf-8")

    expire = GetExpirationTimeString()
    data = {
        "userName": user,
        "modulus": converted_modulus,
        "exponent": converted_exponent,
        "email": email,
        "expireOn": expire,
    }

    return json.dumps(data)


def UpdateWindowsKeys(old_metadata, metadata_entry):
    """Return updated metadata contents with the new windows-keys entry."""
    # Simply overwrites the "windows-keys" metadata entry. Production code may
    # want to append new lines to the metadata value and remove any expired
    # entries.
    new_metadata = copy.deepcopy(old_metadata)
    new_metadata["items"] = [{"key": "windows-keys", "value": metadata_entry}]
    return new_metadata


def UpdateInstanceMetadata(compute, instance, zone, project, new_metadata):
    """Update the instance metadata."""
    cmd = compute.instances().setMetadata(
        instance=instance, project=project, zone=zone, body=new_metadata
    )
    return cmd.execute()


def GetSerialPortFourOutput(compute, instance, zone, project):
    """Get the output from serial port 4 from the instance."""
    # Encrypted passwords are printed to COM4 on the windows server:
    port = 4
    cmd = compute.instances().getSerialPortOutput(
        instance=instance, project=project, zone=zone, port=port
    )
    output = cmd.execute()
    return output["contents"]


def GetEncryptedPasswordFromSerialPort(serial_port_output, modulus):
    """Find and return the correct encrypted password, based on the modulus."""
    # In production code, this may need to be run multiple times if the output
    # does not yet contain the correct entry.

    converted_modulus = modulus.decode("utf-8")

    output = serial_port_output.split("\n")
    for line in reversed(output):
        try:
            entry = json.loads(line)
            if converted_modulus == entry["modulus"]:
                return entry["encryptedPassword"]
        except ValueError:
            pass


def DecryptPassword(encrypted_password, key):
    """Decrypt a base64 encoded encrypted password using the provided key."""

    decoded_password = base64.b64decode(encrypted_password)
    cipher = PKCS1_OAEP.new(key)
    password = cipher.decrypt(decoded_password)
    return password


def Arguments():
    # Create the parser
    args = argparse.ArgumentParser(description="List the content of a folder")

    # Add the arguments
    args.add_argument(
        "--instance", metavar="instance", type=str, help="compute instance name"
    )

    args.add_argument("--zone", metavar="zone", type=str, help="compute zone")

    args.add_argument("--project", metavar="project", type=str, help="gcp project")

    args.add_argument("--username", metavar="username", type=str, help="username")

    args.add_argument("--email", metavar="email", type=str, help="email")

    # return arguments
    return args.parse_args()


def main():
    config_args = Arguments()

    # Setup
    compute = GetCompute()
    key = GetKey()
    modulus, exponent = GetModulusExponentInBase64(key)

    # Get existing metadata
    instance_ref = GetInstance(
        compute, config_args.instance, config_args.zone, config_args.project
    )
    old_metadata = instance_ref["metadata"]
    # Create and set new metadata
    metadata_entry = GetJsonString(
        config_args.username, modulus, exponent, config_args.email
    )
    new_metadata = UpdateWindowsKeys(old_metadata, metadata_entry)

    # Get Serial output BEFORE the modification
    serial_port_output = GetSerialPortFourOutput(
        compute, config_args.instance, config_args.zone, config_args.project
    )

    UpdateInstanceMetadata(
        compute,
        config_args.instance,
        config_args.zone,
        config_args.project,
        new_metadata,
    )

    # Get and decrypt password from serial port output
    # Monitor changes from output to get the encrypted password as soon as it's generated, will wait for 30 seconds
    i = 0
    new_serial_port_output = serial_port_output
    while i <= 20 and serial_port_output == new_serial_port_output:
        new_serial_port_output = GetSerialPortFourOutput(
            compute, config_args.instance, config_args.zone, config_args.project
        )
        i += 1
        time.sleep(3)

    enc_password = GetEncryptedPasswordFromSerialPort(new_serial_port_output, modulus)

    password = DecryptPassword(enc_password, key)
    converted_password = password.decode("utf-8")

    # Display only the password
    print(format(converted_password))


if __name__ == "__main__":
    main()



"""Functional tests."""
import os
import pathlib
import subprocess

from molecule import logger
from molecule.test.conftest import change_dir_to
from molecule.util import run_command

from molecule_plugins.podman import __file__ as module_file

LOG = logger.get_logger(__name__)


def format_result(result: subprocess.CompletedProcess):
    """Return friendly representation of completed process run."""
    return (
        f"RC: {result.returncode}\n"
        + f"STDOUT: {result.stdout}\n"
        + f"STDERR: {result.stderr}"
    )


def test_command_init_scenario(tmp_path: pathlib.Path):
    """Verify that init scenario works."""
    scenario_name = "default"

    with change_dir_to(tmp_path):
        scenario_directory = tmp_path / "molecule" / scenario_name
        cmd = [
            "molecule",
            "init",
            "scenario",
            scenario_name,
            "--driver-name",
            "podman",
        ]
        result = run_command(cmd)
        assert result.returncode == 0

        assert scenario_directory.exists()

        # run molecule reset as this may clean some leftovers from other
        # test runs and also ensure that reset works.
        result = run_command(["molecule", "reset"])  # default sceanario
        assert result.returncode == 0

        result = run_command(["molecule", "reset", "-s", scenario_name])
        assert result.returncode == 0

        cmd = ["molecule", "--debug", "test", "-s", scenario_name]
        result = run_command(cmd)
        assert result.returncode == 0


def test_sample() -> None:
    """Runs the sample scenario present at the repository root."""
    result = run_command(["molecule", "test", "-s", "test-podman"])  # default sceanario
    assert result.returncode == 0


def test_dockerfile():
    """Verify that our embedded dockerfile can be build."""
    result = subprocess.run(
        ["ansible-playbook", "--version"],
        check=False,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.DEVNULL,
        shell=False,
        universal_newlines=True,
    )
    assert result.returncode == 0, result
    assert "ansible-playbook" in result.stdout

    module_path = os.path.dirname(module_file)
    assert os.path.isdir(module_path)
    env = os.environ.copy()
    env["ANSIBLE_FORCE_COLOR"] = "0"
    result = subprocess.run(
        ["ansible-playbook", "-i", "localhost,", "playbooks/validate-dockerfile.yml"],
        check=False,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.DEVNULL,
        shell=False,
        cwd=module_path,
        universal_newlines=True,
        env=env,
    )
    assert result.returncode == 0, format_result(result)
    # , result



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Docker Driver Module."""

from __future__ import absolute_import

import os
from typing import Dict

from molecule import logger
from molecule.api import Driver
from molecule.util import sysexit_with_message

log = logger.get_logger(__name__)


class Docker(Driver):
    """
    Docker Driver Class.

    The class responsible for managing `Docker`_ containers.  `Docker`_ is
    the default driver used in Molecule.

    Molecule leverages Ansible's `docker_container`_ module, by mapping
    variables from ``molecule.yml`` into ``create.yml`` and ``destroy.yml``.

    Molecule leverages Ansible's `docker_network`_ module, by mapping variable
    ``docker_networks`` into ``create.yml`` and ``destroy.yml``.

    .. _`docker_container`: https://docs.ansible.com/ansible/latest/modules/docker_container_module.html
    .. _`docker_network`: https://docs.ansible.com/ansible/latest/modules/docker_network_module.html
    .. _`Docker Security Configuration`: https://docs.docker.com/engine/reference/run/#security-configuration
    .. _`Docker daemon socket options`: https://docs.docker.com/engine/reference/commandline/dockerd/#daemon-socket-option

    .. code-block:: yaml

        driver:
          name: docker
        platforms:
          - name: instance
            hostname: instance
            image: image_name:tag
            dockerfile: Dockerfile.j2
            pull: True|False
            pre_build_image: True|False
            registry:
              url: registry.example.com
              credentials:
                username: $USERNAME
                password: $PASSWORD
                email: user@example.com
                user: root
            override_command: True|False
            command: sleep infinity
            tty: True|False
            pid_mode: host
            privileged: True|False
            security_opts:
              - seccomp=unconfined
            cgroupns_mode: host|private
            devices:
              - /dev/fuse:/dev/fuse:rwm
            volumes:
              - /sys/fs/cgroup:/sys/fs/cgroup:ro
            keep_volumes: True|False
            tmpfs:
              - /tmp
              - /run
            capabilities:
              - SYS_ADMIN
            sysctls:
              net.core.somaxconn: 1024
              net.ipv4.tcp_syncookies: 0
            exposed_ports:
              - 53/udp
              - 53/tcp
            published_ports:
              - 0.0.0.0:8053:53/udp
              - 0.0.0.0:8053:53/tcp
            ulimits:
              - nofile:262144:262144
            dns_servers:
              - 8.8.8.8
            etc_hosts: "{'host1.example.com': '10.3.1.5'}"
            docker_networks:
              - name: foo
                ipam_config:
                  - subnet: '10.3.1.0/24'
                    gateway: '10.3.1.254'
            networks:
              - name: foo
              - name: bar
            network_mode: host
            purge_networks: true
            docker_host: tcp://localhost:12376
            cacert_path: /foo/bar/ca.pem
            cert_path: /foo/bar/cert.pem
            key_path: /foo/bar/key.pem
            tls_verify: true
            env:
              FOO: bar
            restart_policy: on-failure
            restart_retries: 1
            buildargs:
                http_proxy: http://proxy.example.com:8080/

    If specifying the `CMD`_ directive in your ``Dockerfile.j2`` or consuming a
    built image which declares a ``CMD`` directive, then you must set
    ``override_command: False``. Otherwise, Molecule takes care to honour the
    value of the ``command`` key or uses the default of ``bash -c "while true;
    do sleep 10000; done"`` to run the container until it is provisioned.

    When attempting to utilize a container image with `systemd`_ as your init
    system inside the container to simulate a real machine, make sure to set
    the ``privileged``, ``volumes``, ``command``, and ``env``
    values. An example using the ``centos:8`` image is below:

    .. note:: Do note that running containers in privileged mode is considerably
              less secure. For details, please reference `Docker Security
              Configuration`_

    .. note:: On some systems (macOS) you might have to set ``cgroupns_mode`` to
              ``host`` for `systemd` to work. See `Docker Desktop release
              notes`_ for more information.

    .. _`Docker Desktop release notes`: https://docs.docker.com/desktop/release-notes/#bug-fixes-and-minor-changes-19

    .. note:: With the environment variable ``DOCKER_HOST`` the user can bind
              Molecule to a different `Docker`_ socket than the default
              ``unix:///var/run/docker.sock``. ``tcp``, ``fd`` and ``ssh``
              socket types can be configured. For details, please reference
              `Docker daemon socket options`_.

    .. code-block:: yaml

        platforms:
        - name: instance
          image: quay.io/centos/centos:stream8
          privileged: true
          volumes:
            - "/sys/fs/cgroup:/sys/fs/cgroup:rw"
          command: "/usr/sbin/init"
          tty: True
          env:
            container: docker

    .. code-block:: bash

        $ python3 -m pip install molecule[docker]

    When pulling from a private registry, it is the user's discretion to decide
    whether to use hard-code strings or environment variables for passing
    credentials to molecule.

    .. important::

        Hard-coded credentials in ``molecule.yml`` should be avoided, instead use
        `variable substitution`_.

    Provide a list of files Molecule will preserve, relative to the scenario
    ephemeral directory, after any ``destroy`` subcommand execution.

    .. code-block:: yaml

        driver:
          name: docker
          safe_files:
            - foo

    .. _`Docker`: https://www.docker.com
    .. _`systemd`: https://www.freedesktop.org/wiki/Software/systemd/
    .. _`CMD`: https://docs.docker.com/engine/reference/builder/#cmd
    """  # noqa

    _passed_sanity = False

    def __init__(self, config=None):
        """Construct Docker."""
        super(Docker, self).__init__(config)
        self._name = "docker"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def login_cmd_template(self):
        return (
            "docker exec "
            "-e COLUMNS={columns} "
            "-e LINES={lines} "
            "-e TERM=bash "
            "-e TERM=xterm "
            "-ti {instance} bash"
        )

    @property
    def default_safe_files(self):
        return [os.path.join(self._config.scenario.ephemeral_directory, "Dockerfile")]

    @property
    def default_ssh_connection_options(self):
        return []

    def login_options(self, instance_name):
        return {"instance": instance_name}

    def ansible_connection_options(self, instance_name):
        x = {"ansible_connection": "community.docker.docker"}
        if "DOCKER_HOST" in os.environ:
            x["ansible_docker_extra_args"] = f"-H={os.environ['DOCKER_HOST']}"
        return x

    def sanity_checks(self):
        """Implement Docker driver sanity checks."""
        if self._passed_sanity:
            return

        log.info("Sanity checks: '%s'", self._name)
        try:
            import docker
            import requests

            docker_client = docker.from_env()
            docker_client.ping()
        except requests.exceptions.ConnectionError:
            msg = (
                "Unable to contact the Docker daemon. "
                "Please refer to https://docs.docker.com/config/daemon/ "
                "for managing the daemon"
            )
            sysexit_with_message(msg)

        self._passed_sanity = True

    def reset(self):
        import docker

        client = docker.from_env()
        for c in client.containers.list(filters={"label": "owner=molecule"}):
            log.info("Stopping docker container %s ...", c.id)
            c.stop(timeout=3)
        result = client.containers.prune(filters={"label": "owner=molecule"})
        for c in result.get("ContainersDeleted") or []:
            log.info("Deleted container %s", c)
        for n in client.networks.list(filters={"label": "owner=molecule"}):
            log.info("Removing docker network %s ...", n.name)
            n.remove()

    @property
    def required_collections(self) -> Dict[str, str]:
        """Return collections dict containing names and versions required."""
        # https://galaxy.ansible.com/community/docker
        return {"community.docker": "3.0.2", "ansible.posix": "1.4.0"}
